@@ -6,7 +6,8 @@
 # Authors: Hans Petter Jansson <hpj@ximian.com>,
 #          Arturo Espinosa <arturo@ximian.com>,
 #          Tambet Ingo <tambet@ximian.com>.
-#          Grzegorz Golawski <grzegol@pld-linux.org> (PLD Support)
+#          Grzegorz Golawski <grzegol@pld-linux.org> (PLD Support),
+#          Milan Bouchet-Valat <nalimilan@club.fr>.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Library General Public License as published
@@ -45,6 +46,8 @@
 
 package Users::Users;
 
+use Authen::PAM;
+
 use Utils::Util;
 use Utils::Report;
 use Utils::File;
@@ -72,15 +75,23 @@ $cmd_deluser  = &Utils::File::locate_tool ("deluser");
 $cmd_chfn     = &Utils::File::locate_tool ("chfn");
 $cmd_pw       = &Utils::File::locate_tool ("pw");
 
+$cmd_passwd   = &Utils::File::locate_tool ("passwd");
+$cmd_chpasswd = &Utils::File::locate_tool ("chpasswd");
+
 # enum like for verbose group array positions
 my $i = 0;
-my $LOGIN   = $i++;
-my $PASSWD  = $i++;
-my $UID     = $i++;
-my $GID     = $i++;
-my $COMMENT = $i++;
-my $HOME    = $i++;
-my $SHELL   = $i++;
+my $LOGIN         = $i++;
+my $PASSWD        = $i++;
+my $UID           = $i++;
+my $GID           = $i++;
+my $COMMENT       = $i++;
+my $HOME          = $i++;
+my $SHELL         = $i++;
+my $PASSWD_STATUS = $i++;
+my $ENC_HOME      = $i++;
+my $HOME_FLAGS    = $i++;
+my $LOCALE        = $i++;
+my $LOCATION      = $i++;
 
 %login_defs_prop_map = ();
 %profiles_prop_map = ();
@@ -236,7 +247,7 @@ my $logindefs_dist_map = {
   'mandrake-9.0'    => $rh_logindefs_defaults,
   'pld-1.0'         => $rh_logindefs_defaults,
   'fedora-1'        => $rh_logindefs_defaults,
-  'debian-3.0'      => $rh_logindefs_defaults,
+  'debian'          => $rh_logindefs_defaults,
   'vine-3.0'        => $rh_logindefs_defaults,
   'gentoo'	        => $gentoo_logindefs_defaults,
   'archlinux'       => $gentoo_logindefs_defaults,
@@ -267,45 +278,6 @@ my $logindefs_dist_map = {
   'users_write_groups_fail'       => ['warn', 'Writing groups failed.'],
 });
 
-
-sub do_get_use_md5
-{
-  my ($file) = @_;
-  my ($fh, @line, $i, $use_md5);
-
-  my $fh = &Utils::File::open_read_from_names ("/etc/pam.d/$file");
-  return 0 if (!$fh);
-
-  $use_md5 = 0;
-
-  while (<$fh>)
-  {
-    next if &Utils::Util::ignore_line ($_);
-    chomp;
-    @line = split /[ \t]+/;
-
-    if ($line[0] eq "\@include")
-    {
-      $use_md5 = &do_get_use_md5 ($line[1]);
-    }
-    elsif ($line[0] eq "password")
-    {
-      foreach $i (@line)
-      {
-        $use_md5 = 1 if ($i eq "md5");
-      }
-    }
-  }
-
-  close $fh;
-  return $use_md5;
-}
-
-sub get_use_md5
-{
-  return &do_get_use_md5 ("passwd");
-}
-
 sub logindefs_add_defaults
 {
   # Common for all distros
@@ -396,36 +368,23 @@ sub get
     push @comment, "" while (scalar (@comment) < 5);
     $line[$COMMENT] = [@comment];
 
+    # always return empty string - anyway, passwd should be in /etc/shadow
+    $line[$PASSWD] = "";
+
     $users_hash{$login} = [@line];
+
+    # max value for an unsigned 32 bits integer means no main group
+    $users_hash{$login}[$GID] = 0xFFFFFFFF if (!$users_hash{$login}[$GID]);
+
+    # TODO: read actual values
+    $users_hash{$login}[$PASSWD_STATUS] = 0;
+    $users_hash{$login}[$ENC_HOME] = 0;
+    $users_hash{$login}[$HOME_FLAGS] = 0;
+    $users_hash{$login}[$LOCALE] = "";
+    $users_hash{$login}[$LOCATION] = "";
   }
 
   &Utils::File::close_file ($ifh);
-  $ifh = &Utils::File::open_read_from_names(@shadow_names);
-
-  if ($ifh)
-  {
-    while (<$ifh>)
-    {
-      chomp;
-
-      # FreeBSD allows comments in the shadow passwd file.
-      next if &Utils::Util::ignore_line ($_);
-
-      @line = split ':', $_, -1;
-      $login = shift @line;
-      $passwd = shift @line;
-
-      # do not add data if the user isn't present
-      next if (!exists $users_hash{$login});
-
-      $users_hash{$login}[$PASSWD] = $passwd;
-
-      # FIXME: add the rest of the fields?
-      #push @{$$users_hash{$login}}, @line;
-    }
-
-    &Utils::File::close_file ($ifh);
-  }
 
   # transform the hash into an array
   foreach $login (keys %users_hash)
@@ -438,33 +397,53 @@ sub get
 
 sub del_user
 {
-	my ($user) = @_;
-  my ($command);
+  my ($user) = @_;
+  my (@command, $remove_home);
+  
+  $remove_home = $$user[$HOME_FLAGS] & 1;
 	
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    $command = "$cmd_pw userdel -n \'" . $$user[$LOGIN] . "\' ";
-  }
-  else
-  {
-    if ($cmd_deluser)
-    {
-      $command = "$cmd_deluser '". $$user[$LOGIN] . "'";
-    }
+    if ($remove_home)
+      {
+        @command = ($cmd_pw, "userdel", "-r", "-n", $$user[$LOGIN]);
+      }
     else
+      {
+        @command = ($cmd_pw, "userdel", "-n", $$user[$LOGIN]);
+      }
+  }
+  elsif ($cmd_deluser) # use deluser (preferred method)
     {
-      $command = "$cmd_userdel \'" . $$user[$LOGIN] . "\'";
+      if ($remove_home)
+      {
+        @command = ($cmd_deluser, "--remove-home", $$user[$LOGIN]);
+      }
+      else
+      {
+        @command = ($cmd_deluser, $$user[$LOGIN]);
+      }
     }
+  else # use userdel
+    {
+      if ($remove_home)
+      {
+        @command = ($cmd_userdel, "--remove", $$user[$LOGIN]);
+      }
+      else
+      {
+        @command = ($cmd_userdel, $$user[$LOGIN]);
+      }
   }
 
-  &Utils::File::run ($command);
+  &Utils::File::run (@command);
 }
 
 sub change_user_chfn
 {
   my ($login, $old_comment, $comment) = @_;
   my ($fname, $office, $office_phone, $home_phone);
-  my ($command, $str);
+  my (@command, $str);
 
   return if !$login;
 
@@ -474,14 +453,15 @@ sub change_user_chfn
 
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    $command = "$cmd_pw usermod -n " . $login . " -c \'" . $str . "\'";
+    @command = ($cmd_pw, "usermod", "-n", $login,
+                                    "-c", $str);
   }
   else
   {
-    $command = "$cmd_usermod -c \'" . $str . "\' " . $login;
+    @command = ($cmd_usermod, "-c", $str, $login);
   }
 
-  &Utils::File::run ($command);
+  &Utils::File::run (@command);
 }
 
 # modifies /etc/shadow directly, not good practice,
@@ -515,79 +495,92 @@ sub modify_shadow_password
 sub set_passwd
 {
   my ($login, $password) = @_;
-  my ($pwdpipe, $command);
+  my ($pwdpipe);
 
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-
-    $command = "$cmd_pw usermod -H 0";
+    my ($command);
+    $command = "$cmd_pw usermod \'$login\' -h 0";
     $pwdpipe = &Utils::File::run_pipe_write ($command);
     print $pwdpipe $password;
     &Utils::File::close_file ($pwdpipe);
   }
   elsif ($Utils::Backend::tool{"system"} eq "SunOS")
   {
-    &modify_shadow_password ($login, $password);
+    my ($command);
+    $command = "$cmd_passwd --stdin \'$login\'";
+    $pwdpipe = &Utils::File::run_pipe_write ($command);
+    print $pwdpipe $password;
+    &Utils::File::close_file ($pwdpipe);
   }
   else
   {
-    $command = "$cmd_usermod " .
-        " -p '" . $password . "' " . $login;
-
-    &Utils::File::run ($command);
+    $pwdpipe = &Utils::File::run_pipe_write ($cmd_chpasswd);
+    print $pwdpipe "$login:$password";
+    &Utils::File::close_file ($pwdpipe);
   }
 }
 
+# This function allows empty values to be passed, in which cas
+# the platform's tools will choose the default.
 sub add_user
 {
-	my ($user) = @_;
-	my ($home_parents, $tool_mkdir);
+  my ($user) = @_;
+  my ($tool_mkdir, $chown_home, $real_uid, $real_gid);
   
   $tool_mkdir = &Utils::File::locate_tool ("mkdir");
 
+  # If directory is specified, ensure its parents exist.
+  # When using default prefix, we assume the directory exists.
+  if ($$user[$HOME])
+  {
+    my ($home_parents);
+
+    $home_parents = $$user[$HOME];
+    $home_parents =~ s/\/+[^\/]+\/*$//;
+    &Utils::File::run ($tool_mkdir, "-p", $home_parents);
+  }
+
+  # max value means default UID or GID here
+  $real_uid = ($$user[$UID] != 0xFFFFFFFF);
+  $real_gid = ($$user[$GID] != 0xFFFFFFFF);
+
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    my $pwdpipe;
-    my $home;
+    my $logindefs;
 
     # FreeBSD doesn't create the home directory
-    $home = $$user[$HOME];
-    &Utils::File::run ("$tool_mkdir -p $home");
+    if (!$$user[$HOME])
+    {
+      $logindefs = &get_logindefs ();
+      $$user[$HOME] = "$$logindefs{'home_prefix'}/$$user[$LOGIN]";
+    }
+    &Utils::File::run ($tool_mkdir, "-p", $$user[$HOME]);
 
-    $command = "$cmd_pw useradd " .
-        " -n \'" . $$user[$LOGIN] . "\'" .
-        " -u \'" . $$user[$UID]   . "\'" .
-        " -d \'" . $$user[$HOME]  . "\'" .
-        " -g \'" . $$user[$GID]   . "\'" .
-        " -s \'" . $$user[$SHELL] . "\'" .
-        " -H 0"; # pw(8) reads password from STDIN
+    @command = ($cmd_pw, "useradd", "-n", $$user[$LOGIN],
+                                    "-h", "-"); # disable login until password is set
 
-    $pwdpipe = &Utils::File::run_pipe_write ($command);
-    print $pwdpipe $$user[$PASSWD];
-    &Utils::File::close_file ($pwdpipe);
+    push (@command, ("-s", $$user[$HOME])) if ($$user[$HOME]);
+    push (@command, ("-s", $$user[$SHELL])) if ($$user[$SHELL]);
+    push (@command, ("-u", $$user[$UID])) if ($real_uid);
+    push (@command, ("-g", $$user[$GID])) if ($real_gid);
+
+    &Utils::File::run (@command);
   }
   elsif ($Utils::Backend::tool{"system"} eq "SunOS")
   {
-    $home_parents = $$user[$HOME];
-    $home_parents =~ s/\/+[^\/]+\/*$//;
-    &Utils::File::run ("$tool_mkdir -p $home_parents");
+    @command = ($cmd_useradd);
 
-    $command = "$cmd_useradd" .
-        " -d \'" . $$user[$HOME]  . "\'" .
-        " -g \'" . $$user[$GID]   . "\'" .
-        " -s \'" . $$user[$SHELL] . "\'" .
-        " -u \'" . $$user[$UID]   . "\'" .
-        " \'"    . $$user[$LOGIN] . "\'";
+    push (@command, ("-d", $$user[$HOME])) if ($$user[$HOME]);
+    push (@command, ("-s", $$user[$SHELL])) if ($$user[$SHELL]);
+    push (@command, ("-u", $$user[$UID])) if ($real_uid);
+    push (@command, ("-g", $$user[$GID])) if ($real_gid);
+    push (@command, $$user[$LOGIN]);
 
-    &Utils::File::run ($command);
-    &modify_shadow_password ($$user[$LOGIN], $$user[$PASSWD]);
+    &Utils::File::run (@command);
   }
   else
   {
-    $home_parents = $$user[$HOME];
-    $home_parents =~ s/\/+[^\/]+\/*$//;
-    &Utils::File::run ("$tool_mkdir -p $home_parents");
-
     if ($cmd_adduser &&
         $Utils::Backend::tool{"platform"} !~ /^slackware/ &&
         $Utils::Backend::tool{"platform"} !~ /^archlinux/ &&
@@ -596,88 +589,77 @@ sub add_user
     {
       # use adduser if available and valid (slackware one is b0rk)
       # set empty gecos fields and password, they will be filled out later
-      $command = "$cmd_adduser --gecos '' --disabled-password" .
-          " --home \'"  . $$user[$HOME]   . "\'" .
-          " --gid \'"   . $$user[$GID]    . "\'" .
-          " --shell \'" . $$user[$SHELL]  . "\'" .
-          " --uid \'"   . $$user[$UID]    . "\'" .
-          " \'"         . $$user[$LOGIN]  . "\'";
+      @command = ($cmd_adduser, "--gecos", "",
+                                "--disabled-password");
 
-      &Utils::File::run ($command);
+      push (@command, ("--home", $$user[$HOME])) if ($$user[$HOME]);
+      push (@command, ("--shell", $$user[$SHELL])) if ($$user[$SHELL]);
+      push (@command, ("--uid", $$user[$UID])) if ($real_uid);
+      push (@command, ("--gid", $$user[$GID])) if ($real_gid);
+      push (@command, $$user[$LOGIN]);
 
-      # password can't be set in non-interactive
-      # mode with adduser, call usermod instead
-      $command = "$cmd_usermod " .
-          " -p '" . $$user[$PASSWD] . "' " . $$user[$LOGIN];
-
-      &Utils::File::run ($command);
+      &Utils::File::run (@command);
     }
     else
     {
       # fallback to useradd
-      $command = "$cmd_useradd -m" .
-          " -d \'" . $$user[$HOME]   . "\'" .
-          " -g \'" . $$user[$GID]    . "\'" .
-          " -p \'" . $$user[$PASSWD] . "\'" .
-          " -s \'" . $$user[$SHELL]  . "\'" .
-          " -u \'" . $$user[$UID]    . "\'" .
-          " \'"    . $$user[$LOGIN]  . "\'";
+      @command = ($cmd_useradd, "-m");
 
-      &Utils::File::run ($command);
+      push (@command, ("-d", $$user[$HOME])) if ($$user[$HOME]);
+      push (@command, ("-s", $$user[$SHELL])) if ($$user[$SHELL]);
+      push (@command, ("-u", $$user[$UID])) if ($real_uid);
+      push (@command, ("-g", $$user[$GID])) if ($real_gid);
+      push (@command, $$user[$LOGIN]);
+
+      &Utils::File::run (@command);
     }
   }
 
+  # ensure user owns its home dir if asked
+  $chown_home = $$user[$HOME_FLAGS] & (1 << 1);
+  if ($chown_home && $$user[$HOME] ne "/")
+  {
+    @command = ("chown", "-R", "$$user[$LOGIN]:", $$user[$HOME]);
+    &Utils::File::run (@command);
+  }
+
   &change_user_chfn ($$user[$LOGIN], undef, $$user[$COMMENT]);
+  &set_passwd ($$user[$LOGIN], $$user[$PASSWD]);
+
+  # Return the new user with default values filled.
+  # Returns NULL if user doesn't exist, which means failure.
+  return &get_user ($$user[$LOGIN]);
 }
 
 sub change_user
 {
   my ($old_user, $new_user) = @_;
-	
+
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    my $pwdpipe;
+    @command = ($cmd_pw, "usermod", $$old_user[$LOGIN],
+                         "-l", $$new_user[$LOGIN],
+                         "-u", $$new_user[$UID],
+                         "-d", $$new_user[$HOME],
+                         "-g", $$new_user[$GID],
+                         "-s", $$new_user[$SHELL]);
 
-    $command = "$cmd_pw usermod \'" . $$old_user[$LOGIN] . "\'" .
-        " -l \'" . $$new_user[$LOGIN] . "\'" .
-        " -u \'" . $$new_user[$UID]   . "\'" .
-        " -d \'" . $$new_user[$HOME]  . "\'" .
-        " -g \'" . $$new_user[$GID]   . "\'" .
-        " -s \'" . $$new_user[$SHELL] . "\'" .
-        " -H 0"; # pw(8) reads password from STDIN
-
-    $pwdpipe = &Utils::File::run_pipe_write ($command);
-    print $pwdpipe $$new_user[$PASSWD];
-    &Utils::File::close_file ($pwdpipe);
-  }
-  elsif ($Utils::Backend::tool{"system"} eq "SunOS")
-  {
-    $command = "$cmd_usermod" .
-        " -d \'" . $$new_user[$HOME]   . "\'" .
-        " -g \'" . $$new_user[$GID]    . "\'" .
-        " -l \'" . $$new_user[$LOGIN]  . "\'" .
-        " -s \'" . $$new_user[$SHELL]  . "\'" .
-        " -u \'" . $$new_user[$UID]    . "\'" .
-        " \'" . $$old_user[$LOGIN] . "\'";
-
-    &Utils::File::run ($command);
-    &modify_shadow_password ($$new_user[$LOGIN], $$new_user[$PASSWD]);
+    &Utils::File::run (@command);
   }
   else
   {
-    $command = "$cmd_usermod" .
-        " -d \'" . $$new_user[$HOME]   . "\'" .
-        " -g \'" . $$new_user[$GID]    . "\'" .
-        " -l \'" . $$new_user[$LOGIN]  . "\'" .
-        " -p \'" . $$new_user[$PASSWD] . "\'" .
-        " -s \'" . $$new_user[$SHELL]  . "\'" .
-        " -u \'" . $$new_user[$UID]    . "\'" .
-        " \'" . $$old_user[$LOGIN] . "\'";
+    @command = ($cmd_usermod, "-d", $$new_user[$HOME],
+                              "-g", $$new_user[$GID],
+                              "-l", $$new_user[$LOGIN],
+                              "-s", $$new_user[$SHELL],
+                              "-u", $$new_user[$UID],
+                                    $$old_user[$LOGIN]);
 
-    &Utils::File::run ($command);
+    &Utils::File::run (@command);
   }
 
   &change_user_chfn ($$new_user[$LOGIN], $$old_user[$COMMENT], $$new_user[$COMMENT]);
+  &set_passwd ($$new_user[$LOGIN], $$new_user[$PASSWD]);
 }
 
 sub set_logindefs
@@ -714,57 +696,7 @@ sub set_logindefs
   }
 }
 
-sub set
-{
-  my ($config) = @_;
-  my ($old_config, %users);
-  my (%config_hash, %old_config_hash);
-
-  if ($config)
-  {
-    # Make backups manually, otherwise they don't get backed up.
-    &Utils::File::do_backup ($_) foreach (@passwd_names);
-    &Utils::File::do_backup ($_) foreach (@shadow_names);
-
-    $old_config = &get ();
-
-    foreach $i (@$config) 
-    {
-      $users{$$i[$LOGIN]} |= 1;
-      $config_hash{$$i[$LOGIN]} = $i;
-    }
-	
-    foreach $i (@$old_config)
-    {
-      $users{$$i[$LOGIN]} |= 2;
-      $old_config_hash{$$i[$LOGIN]} = $i;
-    }
-
-    # Delete all groups that only appeared in the old configuration
-    foreach $i (sort (keys (%users)))
-    {
-      $state = $users{$i};
-
-      if ($state == 1)
-      {
-        # Users with state 1 have been added to the config
-        &add_user ($config_hash{$i});
-      }
-      elsif ($state == 2)
-      {
-        # Users with state 2 have been deleted from the config
-        &del_user ($old_config_hash{$i});
-      }
-      elsif (($state == 3) &&
-             (!Utils::Util::struct_eq ($config_hash{$i}, $old_config_hash{$i})))
-      {
-        &change_user ($old_config_hash{$i}, $config_hash{$i});
-      }
-    }
-  }
-}
-
-sub get_user
+sub get_self
 {
   my ($uid) = @_;
   my ($users) = &get ();
@@ -772,28 +704,66 @@ sub get_user
   foreach $user (@$users)
   {
     next if ($uid != $$user[$UID]);
-    return ($$user[$UID], $$user[$PASSWD], $$user[$COMMENT]);
+    return ($$user[$COMMENT], $$user[$LOCALE]);
   }
 
-  return ($uid, "", []);
+  return ([""], "");
+}
+
+sub get_user
+{
+  my ($login) = @_;
+  my ($users) = &get ();
+
+  foreach $user (@$users)
+  {
+    next if ($login ne $$user[$LOGIN]);
+    return $user;
+  }
+
+  return NULL;
 }
 
 sub set_user
 {
-  my ($uid, $passwd, @comment) = @_;
+  my ($new_user) = @_;
   my ($users) = &get ();
 
+  # Make backups manually, otherwise they don't get backed up.
+  &Utils::File::do_backup ($_) foreach (@passwd_names);
+  &Utils::File::do_backup ($_) foreach (@shadow_names);
+
+  foreach $user (@$users)
+  {
+    if ($$new_user[$UID] == $$user[$UID])
+    {
+      &change_user ($user, $new_user);
+      return;
+    }
+  }
+}
+
+sub set_self
+{
+  my ($uid, $old_passwd, $new_passwd, @comments) = @_;
+  my ($users) = &get ();
+
+  # Make backups manually, otherwise they don't get backed up.
+  &Utils::File::do_backup ($_) foreach (@passwd_names);
+  &Utils::File::do_backup ($_) foreach (@shadow_names);
+
   foreach $user (@$users)
   {
     if ($uid == $$user[$UID])
     {
-      &set_passwd ($$user[$LOGIN], $passwd);
-      &change_user_chfn ($$user[$LOGIN], undef, @comment);
+      &change_user_chfn ($$user[$LOGIN], $$user[$COMMENT], @comments);
       return;
     }
   }
+  # TODO: change password
 }
 
+
 sub get_files
 {
   my ($arr);
